/* Funcoes graficas basicas 16 bits para video */



// calcula incremento de x (inverso da declividade)
#define INCREMENTOX( XI, YI, XF, YF ) ( (float) ( ( XF ) - ( XI ) ) / ( ( YF ) - ( YI ) ) )



/* construtor - Aloca memoria, inicia modo grafico, programa palette,...
**********************************************************/
CGr16::CGr16 ()
{
	// entra em modo grafico
	mModoMCGA ();

   // aponta para area de video
   m_pyVga = (BYTE*) MK_FP ( 0xA000, 0 );

	// faz com que a proporcao entre X e Y dos pixels seja 1
	outportb ( 0x3C2, 0xE3 );

	// programa palette
	for ( INT ncG = 64; ncG < 128; ncG ++ )
		mAtrPalCor ( ncG, 0, 0, ncG );
}



/* destrutor - Volta ao modo texto, desaloca memoria
**********************************************************/
CGr16::~CGr16 ()
{
	// entra em modo texto
	mModoTexto ();
}



/* Inicia modo grafico MCGA: 320x200x256
**********************************************************/
void CGr16::mModoMCGA ()
{
	REGS
		stRegsE, stRegsS;      // registradores

	stRegsE.x.ax = 0x0013;
	int86 ( 0x10, &stRegsE, &stRegsS );
}



/* Inicia modo texto
**********************************************************/
void CGr16::mModoTexto ()
{
	REGS
		stRegsE, stRegsS;      // registradores

	stRegsE.x.ax = 0x0003;
	int86 ( 0x10, &stRegsE, &stRegsS );
}



/* Atribui uma quantidade de RGB para uma cor (entrada) da palette
**********************************************************/
void CGr16::mAtrPalCor ( BYTE yCor, BYTE yR, BYTE yG, BYTE yB )
{
	outportb ( 0x03C8, yCor );
	outportb ( 0x03C9, yR );
	outportb ( 0x03C9, yG );
	outportb ( 0x03C9, yB );
}



/* Atribui toda a tela com uma soh cor.
**********************************************************/
void CGr16::mAtrVdCor ( BYTE yCor )
{
   memset ( (void*) m_pyVga, yCor, 0xffff );
}



/* Marca um pixel na tela
**********************************************************/
void CGr16::mAtrVdPixel ( INT nX, INT nY, BYTE yCor )
{
	memset ( (void*) ( m_pyVga + nX + nY * RESX ), yCor, 1 );
}



/* 1 - Desenha um quadrilatero na memoria virtual de video usando
o incremento de x para calcular posicao de x. Entrada eh vertice por
vertice. Para garantir desenho de poligino convexo os vertices
adjacentes devem ocupar posicoes contiguas nas matrizes de vertices
do objeto quadrilatero. Ou seja, visualmente devem ser definidos
em ordem horaria ou anti-horaria. Exemplo:

								(x4, y4)
								*                -
					        / \                 loop
							 /   \                    |
				(x1, y1) *     \                   |
							 \		 * (x3, y3)         | vai calculando posicoes de x
					        \	/                   | entre retas horizontais e
								\ /                    V retas das bordas do poligono
								 *               -
								 (x2, y2)

Equacao das retas:
Y = aX + b, onde
            a eh declividade
            b eh deslocamento da reta no eixo y
Equacao usada nas interseccoes:
X = ( deltaY * incrementoX ) + Xvertice, onde
     		   incrementoX eh 1 / a
		   	deltaY eh a distancia entre o Y de um vertice da reta e Y do loop
            Xvertice eh o X do vertice usado no calculo de deltaY

Os vertices sao numerados de 1 a 4 correspondendo a sua ordem
na matriz de vertices do objeto poligono.
Sao chamados vertices superior ou inferior os vertices dos extremos
considerando y, no exemplo sao (x4, y4) e (x2, y2), respectivamente.
**********************************************************/
void CGr16::mAtrVdQuad2D ( INT16 nX1, INT16 nY1, INT16 nX2, INT16 nY2,
                           INT16 nX3, INT16 nY3, INT16 nX4, INT16 nY4,
									BYTE yCor )
{
   // recorta poligono, faz com que vertices permanecam dentro da tela
	// verifica se os X estao fora dos limites, se estao os corrige
   // testa x1
	if ( nX1 < MINX )
		nX1 = MINX;
	else
		if ( nX1 > MAXX )
			nX1 = MAXX;
   // testa x2
	if ( nX2 < MINX )
		nX2 = MINX;
	else
		if ( nX2 > MAXX )
			nX2 = MAXX;
   // testa x3
	if ( nX3 < MINX )
		nX3 = MINX;
	else
		if ( nX3 > MAXX )
			nX3 = MAXX;
   // testa x4
	if ( nX4 < MINX )
		nX4 = MINX;
	else
		if ( nX4 > MAXX )
			nX4 = MAXX;
	// verifica se os Y estao fora dos limites, se estao os corrige
   // testa y1
	if ( nY1 < MINY )
		nY1 = MINY;
	else
		if ( nY1 > MAXY )
			nY1 = MAXY;
   // testa y2
	if ( nY2 < MINY )
		nY2 = MINY;
	else
		if ( nY2 > MAXY )
			nY2 = MAXY;
   // testa y3
	if ( nY3 < MINY )
		nY3 = MINY;
	else
		if ( nY3 > MAXY )
			nY3 = MAXY;
   // testa y4
	if ( nY4 < MINY )
		nY4 = MINY;
	else
		if ( nY4 > MAXY )
			nY4 = MAXY;

	// define maior e menor Y (vertices extremos em y).
	INT
		nYSup = nY1,
		nYInf = nY2;
   // se o valor do y superior for mais abaixo, na tela, entao muda
   if ( nYSup > nY1 )
      nYSup = nY1;
   if ( nYSup > nY2 )
      nYSup = nY2;
   if ( nYSup > nY3 )
      nYSup = nY3;
   if ( nYSup > nY4 )
      nYSup = nY4;
   // se o valor do y inferior for mais acima, na tela, entao muda
   if ( nYInf < nY1 )
      nYInf = nY1;
   if ( nYInf < nY2 )
      nYInf = nY2;
   if ( nYInf < nY3 )
      nYInf = nY3;
   if ( nYInf < nY4 )
      nYInf = nY4;

   // define se eh possivel calcular incrementos de x e se possivel calcula
   float
   	fIncX12 = 0,
   	fIncX14 = 0,
   	fIncX23 = 0,
   	fIncX34 = 0;
   BOOL
   	bIncValido12 = FALSE,
   	bIncValido14 = FALSE,
   	bIncValido23 = FALSE,
   	bIncValido34 = FALSE;
   // sao calculados valores apenas entre pontos que nao estao em colunas ==
   if ( nY1 != nY2 ) {
   	fIncX12 = INCREMENTOX ( nX1, nY1, nX2, nY2 );
      bIncValido12 = ! bIncValido12;
   }
   if ( nY1 != nY4 ) {
   	fIncX14 = INCREMENTOX ( nX1, nY1, nX4, nY4 );
      bIncValido14 = ! bIncValido14;
   }
   if ( nY2 != nY3 ) {
   	fIncX23 = INCREMENTOX ( nX2, nY2, nX3, nY3 );
      bIncValido23 = ! bIncValido23;
   }
   if ( nY3 != nY4 ) {
   	fIncX34 = INCREMENTOX ( nX3, nY3, nX4, nY4 );
      bIncValido34 = ! bIncValido34;
   }

   // varre poligono do extremo superior ateh o extremo inferior
   INT16
   	nXIni,    // deve possuir apenas x inicial da linha horizontal
      nXFin,    // deve possuir apenas x final da linha horizontal
      nX;       // valor x da interceccao entre duas retas
	for ( INT16 ncY = nYSup; ncY < nYInf; ncY ++ ) {

      // valores extremos iniciais de x
   	nXIni = MAXX + 1;
   	nXFin = MINX - 1;

      // se estao sendo definidas posicoes de x entre vertices 1 e 4
      if ( bIncValido14 &&
		     ( ncY >= nY1 && ncY <= nY4 ) ||
		     ( ncY <= nY1 && ncY >= nY4 ) ) {

			nX = ( (float) ( ncY - nY4 ) * fIncX14 ) + nX4;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 1 e 2
      if ( bIncValido12 &&
		     ( ncY >= nY1 && ncY <= nY2 ) ||
		     ( ncY <= nY1 && ncY >= nY2 ) ) {

			nX = ( (float) ( ncY - nY2 ) * fIncX12 ) + nX2;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 2 e 3
      if ( bIncValido23 &&
		     ( ncY >= nY2 && ncY <= nY3 ) ||
		     ( ncY <= nY2 && ncY >= nY3 ) ){

			nX = ( (float) ( ncY - nY3 ) * fIncX23 ) + nX3;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

      // se estao sendo definidas posicoes de x entre vertices 3 e 4
      if ( bIncValido34 &&
		     ( ncY >= nY3 && ncY <= nY4 ) ||
		     ( ncY <= nY3 && ncY >= nY4 ) ) {

			nX = ( (float) ( ncY - nY4 ) * fIncX34 ) + nX4;

         // verifica se eh para alterar x inicial
			if ( nX < nXIni )
				nXIni = nX;
	      // verifica se eh para alterar x final
			if ( nX > nXFin )
				nXFin = nX;
      }

		// verifica se eh para trocar x inicial e final, se sim troca
		if ( nXIni > nXFin ) {

			INT16
				nX = nXIni;
			nXIni = nXFin;
			nXFin = nX;
		}

      // faz reta horizontal se foram definidos valores acima
		if ( nXIni <= nXFin )
			memset ( (void*) ( m_pyVga + RESX * ncY + nXIni ), yCor, nXFin - nXIni );
      else {
      	printf ( MSG_ERROCALCX );
         getch ();
      }
	}
}



/* 2
**********************************************************/
void CGr16::mAtrVdQuad2D ( CQuad2D &oQuad2D, BYTE yCor )
{
	mAtrVdQuad2D ( oQuad2D.m_anX [ 0 ], oQuad2D.m_anY [ 0 ],
                  oQuad2D.m_anX [ 1 ], oQuad2D.m_anY [ 1 ],
                  oQuad2D.m_anX [ 2 ], oQuad2D.m_anY [ 2 ],
                  oQuad2D.m_anX [ 3 ], oQuad2D.m_anY [ 3 ],
                  yCor );
}
